"""
gnusocial.groups
~~~~~~~~~~~~~~~~

Module with group resources.
"""
from .utils import (_post_request, _get_request, docstring,
                    _check_user_target, _check_group_id_and_name,
                    _check_id_and_nickname)
from .docs import (_SERVER_URL_DOC, _USERNAME_DOC, _PASSWORD_DOC, _USER_DICT,
                   _USER_ID_DOC, _SCREEN_NAME_DOC, _GROUP_COUNT, _STATUS_DICT,
                   _GROUP_ID_DOC, _GROUP_NAME_DOC, _MAX_ID_DOC, _SINCE_ID_DOC,
                   _GROUP_DICT, _GROUP_DESCRIPTION, _GROUP_LOCATION,
                   _ALIASES_DOC, _HOMEPAGE_DOC, _FULLNAME_DOC, _USERS_COUNT)


def _resource_path(resource_path: str, **kwargs) -> str:
    _check_id_and_nickname(**kwargs)
    group_id = kwargs.get('id')
    group_name = kwargs.get('nickname')
    if group_id:
        resource_path += '/%d' % group_id
    if group_name:
        resource_path += '/%s' % group_name
    return resource_path


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           id=_GROUP_ID_DOC,
           nickname=_GROUP_NAME_DOC,
           count=_GROUP_COUNT,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           status_dict=_STATUS_DICT)
def timeline(server_url: str,
             username: str='',
             password: str='',
             **kwargs) -> list:
    """Shows a group's timeline. Similar to other timeline resources.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int id: (optional) {id}
:param str nickname: (optional) {nickname}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:rtype: list
:return: list of dicts with following sctructure:

::

    {status_dict}
    """
    resource_path = _resource_path('statusnet/groups/timeline', **kwargs)
    return _post_request(server_url=server_url,
                         resource_path=resource_path,
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           id=_GROUP_ID_DOC,
           nickname=_GROUP_NAME_DOC,
           group_dict=_GROUP_DICT)
def show(server_url: str,
         username: str='',
         password: str='',
         **kwargs) -> dict:
    """Show a group's profile.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int id: (optional) {id}
:param str nickname: (optional) {nickname}
:rtype: dict
:return: dict with following sctructure:

::

    {group_dict}
    """
    resource_path = _resource_path('statusnet/groups/show', **kwargs)
    return _get_request(server_url=server_url,
                        resource_path=resource_path,
                        username=username,
                        password=password,
                        params=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           full_name=_FULLNAME_DOC,
           homepage=_HOMEPAGE_DOC,
           description=_GROUP_DESCRIPTION,
           location=_GROUP_LOCATION,
           aliases=_ALIASES_DOC,
           group_dict=_GROUP_DICT)
def create(server_url: str,
           username: str,
           password: str,
           **kwargs) -> dict:
    """Create a new group.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param str nickname: the name of the new group
:param str full_name: (optional) {full_name}
:param str homepage: (optional) {homepage}
:param str description: (optional) {description}
:param str location: (optional) {location}
:param aliases: (optional) {aliases}
:type aliases: list of strs
:rtype: dict
:return: dict with following sctructure:

::

    {group_dict}
    """
    if 'nickname' not in kwargs:
        raise Exception(
            'You must specify the nickname of the new group.'
        )
    return _post_request(server_url=server_url,
                         resource_path='statusnet/groups/create',
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           id=_GROUP_ID_DOC,
           nickname=_GROUP_NAME_DOC,
           group_dict=_GROUP_DICT)
def join(server_url: str,
         username: str,
         password: str,
         **kwargs) -> dict:
    """Join a group.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param int id: (optional) {id}
:param str nickname: (optional) {nickname}
:rtype: dict
:return: dict with following sctructure:

::

    {group_dict}
    """
    resource_path = _resource_path('statusnet/groups/join', **kwargs)
    return _post_request(server_url=server_url,
                         resource_path=resource_path,
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           id=_GROUP_ID_DOC,
           nickname=_GROUP_NAME_DOC,
           group_dict=_GROUP_DICT)
def leave(server_url: str,
          username: str,
          password: str,
          **kwargs) -> dict:
    """Leave a group.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param int id: (optional) {id}
:param str nickname: (optional) {nickname}
:rtype: dict
:return: dict with following sctructure:

::

    {group_dict}
    """
    resource_path = _resource_path('statusnet/groups/leave', **kwargs)
    return _post_request(server_url=server_url,
                         resource_path=resource_path,
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           count=_GROUP_COUNT,
           group_dict=_GROUP_DICT)
def list_all(server_url: str,
             username: str='',
             password: str='',
             **kwargs) -> list:
    """List all local groups.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int count: (optional) {count}
:rtype: list
:return: list of dicts with following sctructure:

::

    {group_dict}
    """
    return _get_request(server_url=server_url,
                        resource_path='statusnet/groups/list_all',
                        username=username,
                        password=password,
                        params=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           count=_GROUP_COUNT,
           group_dict=_GROUP_DICT)
def user_groups(server_url: str,
                username: str='',
                password: str='',
                **kwargs) -> list:
    """Show the groups a given user is a member of.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:param int count: (optional) {count}
:rtype: list
:return: list of dicts with following sctructure:

::

    {group_dict}
    """
    _check_user_target(username, **kwargs)
    return _post_request(server_url=server_url,
                         resource_path='statusnet/groups/list',
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           id=_GROUP_ID_DOC,
           nickname=_GROUP_NAME_DOC,
           count=_USERS_COUNT,
           user_dict=_USER_DICT)
def members(server_url: str,
            username: str='',
            password: str='',
            **kwargs) -> list:
    """List the members of a given group.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int id: (optional) {id}
:param str nickname: (optional) {nickname}
:param int count: (optional) {count}
:rtype: list
:return: list of dicts with following sctructure:

::

    {user_dict}
    """
    resource_path = _resource_path('statusnet/groups/membership', **kwargs)
    return _post_request(server_url=server_url,
                         resource_path=resource_path,
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           group_id=_GROUP_ID_DOC,
           group_name=_GROUP_NAME_DOC)
def is_member(server_url: str,
              username: str='',
              password: str='',
              **kwargs) -> dict:
    """Determine whether a given user is a member of a given group.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:param int group_id: (optional) {group_id}
:param str group_name: (optional) {group_name}
:rtype: dict
:return: a dict with following structure:

::

    {{
        'is_member': bool
    }}
    """
    _check_user_target(username, **kwargs)
    _check_group_id_and_name(**kwargs)
    return _post_request(server_url=server_url,
                         resource_path='statusnet/groups/is_member',
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           id=_GROUP_ID_DOC,
           nickname=_GROUP_NAME_DOC,
           count=_USERS_COUNT,
           user_dict=_USER_DICT)
def admins(server_url: str,
           username: str='',
           password: str='',
           **kwargs) -> list:
    """List the admins of a given group.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int id: (optional) {id}
:param str nickname: (optional) {nickname}
:param int count: (optional) {count}
:rtype: list
:return: list of dicts with following sctructure:

::

    {user_dict}
    """
    resource_path = _resource_path('statusnet/groups/admins', **kwargs)
    return _get_request(server_url=server_url,
                        resource_path=resource_path,
                        username=username,
                        password=password,
                        params=kwargs).json()



"""
gnusocial.users
~~~~~~~~~~~~~~~

Module with user resources.
"""
from .utils import _get_request, _check_user_target, docstring
from .docs import (_SERVER_URL_DOC, _USERNAME_DOC, _PASSWORD_DOC, _USER_DICT,
                   _USER_ID_DOC, _SCREEN_NAME_DOC, _USERS_COUNT)


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           count=_USERS_COUNT,
           user_dict=_USER_DICT)
def following(server_url: str,
              username: str='',
              password: str='',
              **kwargs) -> list:
    """Returns a collection of user objects for users followed by
    the specified user.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:param int count: (optional) {count}
:rtype: list
:return: a list of dicts with following sctructure:

::

    {user_dict}
    """
    _check_user_target(username, **kwargs)
    return _get_request(server_url=server_url,
                        resource_path='statuses/friends',
                        username=username,
                        password=password,
                        params=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           count=_USERS_COUNT,
           user_dict=_USER_DICT)
def followers(server_url: str,
              username: str='',
              password: str='',
              **kwargs) -> list:
    """Returns a collection of user objects for users following
    the specified user.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:param int count: (optional) {count}
:rtype: list
:return: a list of dicts with following sctructure:

::

    {user_dict}
    """
    _check_user_target(username, **kwargs)
    return _get_request(server_url=server_url,
                        resource_path='statuses/followers',
                        username=username,
                        password=password,
                        params=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           user_dict=_USER_DICT)
def show(server_url: str,
         username: str='',
         password: str='',
         **kwargs):
    """Returns a variety of information about the user specified by the
    required user_id or screen_name parameter. The author’s most recent
    status will be returned inline when possible.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:rtype: dict
:return: dict with following sctructure:

::

    {user_dict}
    """
    _check_user_target(username, **kwargs)
    return _get_request(server_url=server_url,
                        resource_path='users/show',
                        username=username,
                        password=password,
                        params=kwargs).json()



"""
gnusocial.utils
~~~~~~~~~~~~~~~

Module with various utility functions and exception classes.
"""
import re
from typing import Callable
from functools import partial
import requests
from requests.auth import HTTPBasicAuth
from .docs import _SERVER_URL_DOC, _CONFIG_DICT

DOMAIN_REGEX = re.compile(r"http(s|)://(www\.|)(.+?)(/.*|)$")


def docstring(*args, **kwargs):
    """Decorator function to fill in docstring templates."""
    def _wrap(obj):
        obj.__doc__ = obj.__doc__.format(*args, **kwargs)
        return obj
    return _wrap


class ServerURLError(Exception):
    """Exception class for errors in server URL.

:param str server_url: URL of the server
    """
    def __init__(self, server_url: str) -> None:
        self.server_url = server_url
        super().__init__(Exception)

    def __repr__(self) -> str:
        return 'ServerURLError(%r)' % self.server_url

    def __str__(self) -> str:
        return 'Invalid server URL %s' % self.server_url


class AuthenticationError(Exception):
    """Exception class for authentication errors.

:param str server_url: URL of the server
:param str username: name of the authenticating user
:param str password: password of the authenticating user
    """
    def __init__(self, server_url: str, username: str, password: str) -> None:
        self.server_url = server_url
        self.username = username
        self.password = password
        super().__init__(self)

    def __repr__(self) -> str:
        return 'AuthenticationError(%r, %r, %r)' % (self.server_url,
                                                    self.username,
                                                    self.password)

    def __str__(self) -> str:
        return 'Invalid credentials %s:%s for %s' % (self.username,
                                                     self.password,
                                                     self.server_url)


def _check_auth_error(response: requests.models.Response,
                      server_url: str,
                      username: str,
                      password: str) -> None:
    if response.status_code == 401:
        raise AuthenticationError(server_url, username, password)


def _api_path(server_url: str) -> str:
    _validate_server_url(server_url)
    if server_url[-1] != '/':
        server_url += '/'
    return server_url + 'api/'


def _validate_server_url(server_url: str) -> None:
    if not DOMAIN_REGEX.match(server_url):
        raise ServerURLError(server_url)


def _check_connection(server_url: str) -> None:
    response = _get_request(server_url, 'help/test').json()
    if response != 'ok':
        raise requests.ConnectionError(server_url)


def _resource_url(server_url: str,
                  resource_path: str,
                  extension: str='.json') -> str:
    return _api_path(server_url) + resource_path + extension


def _request(request_func: Callable,
             server_url: str,
             resource_path: str,
             username: str='',
             password: str='',
             **kwargs) -> requests.models.Response:
    extension = kwargs.get('extension', '.json')
    req = partial(request_func,
                  url=_resource_url(server_url, resource_path, extension),
                  data=kwargs.get('data'),
                  files=kwargs.get('media'),
                  params=kwargs.get('params'))
    response = None
    if username:
        response = req(auth=HTTPBasicAuth(username, password))
        _check_auth_error(response, server_url, username, password)
    else:
        response = req()
    return response


def _get_request(server_url: str,
                 resource_path: str,
                 username: str='',
                 password: str='',
                 **kwargs) -> requests.models.Response:
    return _request(request_func=requests.get,
                    server_url=server_url,
                    resource_path=resource_path,
                    username=username,
                    password=password,
                    **kwargs)


def _post_request(server_url: str,
                  resource_path: str,
                  username: str='',
                  password: str='',
                  **kwargs) -> requests.models.Response:
    return _request(request_func=requests.post,
                    server_url=server_url,
                    resource_path=resource_path,
                    username=username,
                    password=password,
                    **kwargs)


@docstring(server_url=_SERVER_URL_DOC, config_dict=_CONFIG_DICT)
def config(server_url: str) -> dict:
    """Returns server configuration.

:param str server_url: {server_url}
:rtype: dict
:return: dict with following structure:

::

    {config_dict}

    """
    return _get_request(server_url, 'statusnet/config').json()


def _check_user_target(username: str='', **kwargs) -> None:
    both_targets = 'user_id' in kwargs and 'screen_name' in kwargs
    no_targets = 'user_id' not in kwargs and 'screen_name' not in kwargs
    if both_targets:
        raise Exception(
            "You must either specify the user_id or screen_name."
        )
    if no_targets and not username:
        raise Exception(
            "You must either specify the user_id or screen_name or " +
            "username."
        )


def _check_id_and_nickname(**kwargs) -> None:
    has_group_id = 'id' in kwargs
    has_group_name = 'nickname' in kwargs
    if has_group_id == has_group_name:
        raise Exception(
            "You must either specify the id or nickname."
        )


def _check_group_id_and_name(**kwargs) -> None:
    has_group_id = 'group_id' in kwargs
    has_group_name = 'group_name' in kwargs
    if has_group_id == has_group_name:
        raise Exception(
            "You must either specify the group_id or group_name."
        )



"""
gnusocial.media
~~~~~~~~~~~~~~~

Module with media resources.
"""
from typing import Tuple
from xml.etree import ElementTree as ET
from .utils import _post_request, docstring
from .docs import _SERVER_URL_DOC, _USERNAME_DOC, _PASSWORD_DOC


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC)
def upload(server_url: str,
           username: str,
           password: str,
           filename: str) -> Tuple[str, str]:
    """Uploads media to server and returns attachment URL and file URL.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param str filename: name of the file to upload
:rtype: tuple
:return: tuple with attachment URL and file URL respectively.
    """
    media = {'media': open(filename, 'rb')}
    response_xml = _post_request(server_url=server_url,
                                 resource_path='statusnet/media/upload',
                                 extension='',
                                 username=username,
                                 password=password,
                                 media=media).text
    tree = ET.fromstring(response_xml)
    file_url = tree.find('{http://www.w3.org/2005/Atom}link').get('href', '')
    attachment_url = tree.findtext('mediaurl')
    return (attachment_url, file_url)



"""
gnusocial.search
~~~~~~~~~~~~~~~~

Module with search resources.
"""
from .utils import _get_request, docstring
from .docs import _SERVER_URL_DOC, _USERNAME_DOC, _PASSWORD_DOC, _STATUS_DICT


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           status_dict=_STATUS_DICT)
def search(server_url: str,
           query: str,
           username: str='',
           password: str='') -> list:
    """Returns a collection of statuses matching a specified query.

:param str server_url: {server_url}
:param str query: UTF-8 encoded query
:param str username: (optional) {username}
:param str password: (optional) {password}
:rtype: list
:return: list of dicts with following structure:

::

    {status_dict}
    """
    return _get_request(server_url=server_url,
                        resource_path='search.json?q=%s' % query,
                        username=username,
                        password=password,
                        extension='').json()



"""
gnusocial.accounts
~~~~~~~~~~~~~~~~~~

Module with account resources.
"""
from .utils import _get_request, _post_request, docstring
from .docs import (_SERVER_URL_DOC, _USERNAME_DOC, _PASSWORD_DOC,
                   _USER_LOCATION, _USER_DICT, _NAME_DOC, _URL_DOC,
                   _USER_DESCRIPTION, _PROFILE_LINK_COLOR_DOC)


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC)
def verify_credentials(server_url: str, username: str, password: str) -> None:
    """Tests if supplied user credentials are valid.
    If the credentials were valid, return None.
    If the credentials were invalid, raise gnusocial.utils.AuthenticationError.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:raises gnusocial.utils.AuthenticationError: if credentials are invalid
    """
    _get_request(
        server_url=server_url,
        resource_path='account/verify_credentials',
        username=username,
        password=password
    )


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           user_dict=_USER_DICT)
def update_profile_image(server_url: str,
                         username: str,
                         password: str,
                         filename: str) -> dict:
    """Updates the authenticating user's profile image.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param str filename: filename of the new profile picture
:rtype: dict
:return: dict with following structure:

::

    {user_dict}
    """
    image = {'image': open(filename, 'rb')}
    return _post_request(server_url=server_url,
                         resource_path='account/update_profile_image',
                         username=username,
                         password=password,
                         media=image).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           name=_NAME_DOC,
           url=_URL_DOC,
           location=_USER_LOCATION,
           description=_USER_DESCRIPTION,
           profile_link_color=_PROFILE_LINK_COLOR_DOC,
           user_dict=_USER_DICT)
def update_profile(server_url: str,
                   username: str,
                   password: str,
                   **kwargs) -> dict:
    """Sets some values that users are able to set under the "Account" tab of
    their settings page. Only the parameters specified will be updated.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param str name: (optional) {name}
:param str url: (optional) {url}
:param str location: (optional) {location}
:param str description: (optional) {description}
:param str profile_link_color: (optional) {profile_link_color}
:rtype: dict
:return: dict with following structure:

::

    {user_dict}
    """
    return _post_request(server_url=server_url,
                         resource_path='account/update_profile',
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           fullname=_NAME_DOC,
           homepage=_URL_DOC,
           location=_USER_LOCATION,
           bio=_USER_DESCRIPTION,
           user_dict=_USER_DICT)
def register(server_url: str,
             nickname: str,
             password: str,
             confirm: str,
             **kwargs) -> dict:
    """Registers a new user.

:param str server_url: {server_url}
:param str nickname: name of the new user
:param str password: desired password
:param str confirm: password confirmation
:param str email: (optional) email associated with the new user
:param str fullname: (optional) {fullname}
:param str homepage: (optional) {homepage}
:param str location: (optional) {location}
:param str bio: (optional) {bio}
:rtype dict:
:return: dict with following structure:

::

    {user_dict}
    """
    kwargs['nickname'] = nickname
    kwargs['password'] = password
    kwargs['confirm'] = confirm
    return _post_request(server_url=server_url,
                         resource_path='account/register',
                         data=kwargs).json()



"""
gnusocial.timelines
~~~~~~~~~~~~~~~~~~~

Module with timeline resources
"""
from .utils import _get_request, _check_user_target, docstring
from .docs import (_SERVER_URL_DOC, _USERNAME_DOC, _PASSWORD_DOC, _MAX_ID_DOC,
                   _SINCE_ID_DOC, _STATUSES_COUNT, _STATUS_DICT, _USER_ID_DOC,
                   _SCREEN_NAME_DOC)


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_STATUSES_COUNT,
           status_dict=_STATUS_DICT)
def public(server_url: str,
           username: str='',
           password: str='',
           **kwargs) -> list:
    """Returns the most recent notices, including repeats if they exist, from
    non-protected users.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:rtype: list
:return: list of dicts with following structure:

::

    {status_dict}
    """
    return _get_request(server_url=server_url,
                        resource_path='statuses/public_timeline',
                        username=username,
                        password=password,
                        params=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_STATUSES_COUNT,
           status_dict=_STATUS_DICT)
def home(server_url: str, username: str, password: str, **kwargs) -> list:
    """Returns the most recent notices, including repeats if they exist,
    posted by the authenticating user and the users they follow. This is the
    same timeline seen by a user when they login to their instance.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:rtype: list
:return: list of dicts with following structure:

::

    {status_dict}
    """
    return _get_request(server_url=server_url,
                        resource_path='statuses/home_timeline',
                        username=username,
                        password=password,
                        params=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_STATUSES_COUNT,
           status_dict=_STATUS_DICT,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC)
def friends(server_url: str,
            username: str='',
            password: str='',
            **kwargs) -> list:
    """Alias of statuses/home_timeline for the specified user.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:rtype: list
:return: list of dicts with following structure:

::

    {status_dict}
    """
    _check_user_target(username, **kwargs)
    return _get_request(server_url=server_url,
                        resource_path='statuses/friends_timeline',
                        username=username,
                        password=password,
                        params=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_STATUSES_COUNT,
           status_dict=_STATUS_DICT,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC)
def user(server_url: str,
         username: str='',
         password: str='',
         **kwargs) -> list:
    """Returns the most recent notices posted by the authenticating user. It
    is also possible to request another user's timeline by using the
    screen_name or user_id parameter. The other users timeline will only be
    visible if they are not protected, or if the authenticating user's follow
    request was accepted by the protected user.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:rtype: list
:return: list of dicts with following structure:

::

    {status_dict}
    """
    _check_user_target(username, **kwargs)
    return _get_request(server_url=server_url,
                        resource_path='statuses/user_timeline',
                        username=username,
                        password=password,
                        params=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_STATUSES_COUNT,
           status_dict=_STATUS_DICT,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC)
def mentions(server_url: str,
             username: str='',
             password: str='',
             **kwargs) -> list:
    """Returns the most recent mentions (notices containing @username) for
    the authenticating user or specified user.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:rtype: list
:return: list of dicts with following structure:

::

    {status_dict}
    """
    _check_user_target(username, **kwargs)
    return _get_request(server_url=server_url,
                        resource_path='statuses/mentions',
                        username=username,
                        password=password,
                        params=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_STATUSES_COUNT,
           status_dict=_STATUS_DICT,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC)
def replies(server_url: str,
            username: str='',
            password: str='',
            **kwargs) -> list:
    """Alias of statuses/mentions

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:rtype: list
:return: list of dicts with following structure:

::

    {status_dict}
    """
    _check_user_target(username, **kwargs)
    return _get_request(server_url=server_url,
                        resource_path='statuses/replies',
                        username=username,
                        password=password,
                        params=kwargs).json()



"""
gnusocial.statuses
~~~~~~~~~~~~~~~~~~

Module with status resources.
"""
from .utils import _post_request, _get_request, docstring
from .docs import (_SERVER_URL_DOC, _USERNAME_DOC, _PASSWORD_DOC, _STATUS_DICT,
                   _STATUS_ID_DOC)


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           status_dict=_STATUS_DICT)
def update(server_url: str,
           username: str,
           password: str,
           status: str,
           **kwargs) -> dict:
    """Updates the authenticating user's current status.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param str status: The text of your status update.
:param str source: (optional) The name of application you update the status
    from.
:param int in_reply_to_status_id: (optional) The ID of an existing status that
    the update is in reply to.
:param int lat: (optional) The latitude of the location this status refers to.
    This parameter will be ignored unless it is inside the range -90.0 to
    +90.0 (North is positive) inclusive. It will also be ignored if there
    isn’t a corresponding long parameter.
:param int long: (optional) The longitude of the location this tweet refers to.
    The valid ranges for longitude is -180.0 to +180.0 (East is positive)
    inclusive. This parameter will be ignored if outside that range, if
    it is not a number or if there not a corresponding lat parameter.
:param str media: (optional) the name of the file to upload with the status.
:rtype: dict
:return: dict with following structure:

::

    {status_dict}
    """
    media = None
    if 'media' in kwargs:
        media = {'media': open(kwargs['media'], 'rb').read()}
    kwargs['status'] = status
    return _post_request(server_url=server_url,
                         resource_path='statuses/update',
                         username=username,
                         password=password,
                         data=kwargs,
                         media=media).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           status_id=_STATUS_ID_DOC,
           status_dict=_STATUS_DICT)
def show(server_url: str,
         status_id: int,
         username: str='',
         password: str='') -> dict:
    """Returns a single status, specified by the id parameter. The status'
        author will also be embedded within the status.

:param str server_url: {server_url}
:param int status_id: {status_id}
:param str username: (optional) {username}
:param str password: (optional) {password}
:rtype: dict
:return: dict with following structure:

::

    {status_dict}
    """
    return _get_request(server_url=server_url,
                        resource_path='statuses/show/%d' % status_id,
                        username=username,
                        password=password).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           status_id=_STATUS_ID_DOC,
           status_dict=_STATUS_DICT)
def destroy(server_url: str,
            username: str,
            password: str,
            status_id: int) -> dict:
    """Destroys the status specified by the required ID parameter.
        The authenticating user must be the author of the specified status.
        Returns the destroyed status if successful.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param int status_id: {status_id}
:rtype: dict
:return: dict with following structure:

::

    {status_dict}
    """
    return _post_request(server_url=server_url,
                         resource_path='statuses/destroy/%d' % status_id,
                         username=username,
                         password=password).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           status_id=_STATUS_ID_DOC,
           status_dict=_STATUS_DICT)
def repeat(server_url: str,
           username: str,
           password: str,
           status_id: int) -> dict:
    """Repeats a status. Returns the original status with repeat details
        embedded.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param int status_id: {status_id}
:rtype: dict
:return: dict with following structure:

::

    {status_dict}
    """
    return _post_request(server_url=server_url,
                         resource_path='statuses/retweet/%d' % status_id,
                         username=username,
                         password=password).json()



"""
gnusocial.direct_messages
~~~~~~~~~~~~~~~~~~~~~~~~~

Module with direct messages resources.
"""
from .utils import _post_request, _check_user_target, docstring
from .docs import (_SERVER_URL_DOC, _USERNAME_DOC, _PASSWORD_DOC,
                   _SINCE_ID_DOC, _MAX_ID_DOC, _DM_COUNT, _DM_DICT,
                   _USER_ID_DOC, _SCREEN_NAME_DOC, _TEXT_DOC)


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_DM_COUNT,
           dm_dict=_DM_DICT)
def received(server_url: str, username: str, password: str, **kwargs) -> list:
    """Returns the 20 most recent direct messages sent to the authenticating
    user. Includes detailed information about the sender and recipient user.
    You can request up to 200 direct messages per call, and only the most
    recent 200 DMs will be available using this endpoint.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:rtype: list
:return: list of dicts with following structure:

::

    {dm_dict}
    """
    return _post_request(server_url=server_url,
                         resource_path='direct_messages',
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_DM_COUNT,
           dm_dict=_DM_DICT)
def sent(server_url: str, username: str, password: str, **kwargs) -> list:
    """Returns the 20 most recent direct messages sent by the authenticating
    user. Includes detailed information about the sender and recipient user.
    You can request up to 200 direct messages per call, and only the most
    recent 200 DMs will be available using this endpoint.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:rtype: list
:return: list of dicts with following structure:

::

    {dm_dict}
    """
    return _post_request(server_url=server_url,
                         resource_path='direct_messages/sent',
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           text=_TEXT_DOC,
           dm_dict=_DM_DICT)
def new(server_url: str,
        username: str,
        password: str,
        text: str,
        **kwargs) -> dict:
    """ Sends a new direct message to the specified user from
    the authenticating user.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param str text: {text}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:rtype: dict
:return: dict with following structure:

::

    {dm_dict}
    """
    _check_user_target(**kwargs)
    kwargs['text'] = text
    return _post_request(server_url=server_url,
                         resource_path='direct_messages/new',
                         username=username,
                         password=password,
                         data=kwargs).json()



"""
gnusocial.blocks
~~~~~~~~~~~~~~~~

Module with block resources.
"""
from .utils import _post_request, _check_user_target, docstring
from .docs import (_SERVER_URL_DOC, _USERNAME_DOC, _PASSWORD_DOC, _USER_DICT,
                   _USER_ID_DOC, _SCREEN_NAME_DOC)


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           user_dict=_USER_DICT)
def create(server_url: str,
           username: str,
           password: str,
           **kwargs) -> dict:
    """Blocks the specified user from following the authenticating user.
    In addition the blocked user will not show in the authenticating users
    mentions or timeline (unless repeated by another user). If a follow or
    friend relationship exists it is destroyed.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:rtype: dict
:return: dict with following structure:

::

    {user_dict}
    """
    _check_user_target(**kwargs)
    return _post_request(server_url=server_url,
                         resource_path='blocks/create',
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           user_dict=_USER_DICT)
def destroy(server_url: str,
            username: str,
            password: str,
            **kwargs) -> dict:
    """Un-blocks the specified user from following the authenticating user.
     If relationships existed before the block was instated,
     they will not be restored.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:rtype: dict
:return: dict with following structure:

::

    {user_dict}
    """
    _check_user_target(**kwargs)
    return _post_request(server_url=server_url,
                         resource_path='blocks/destroy',
                         username=username,
                         password=password,
                         data=kwargs).json()



"""
gnusocial.friends
~~~~~~~~~~~~~~~~~

Module with friends resources.
"""
from typing import List
from .utils import _post_request, _check_user_target, docstring
from .docs import (_SERVER_URL_DOC, _USERNAME_DOC, _PASSWORD_DOC, _MAX_ID_DOC,
                   _USER_ID_DOC, _SCREEN_NAME_DOC, _USERS_COUNT, _SINCE_ID_DOC)


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           count=_USERS_COUNT,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC)
def friends(server_url: str,
            username: str='',
            password: str='',
            **kwargs) -> List[int]:
    """Returns IDs of users the auntenticated or
    specified user is following (otherwise known as their "friends").

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:param int count: (optional) {count}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:rtype: list
:return: a list of user IDs
    """
    _check_user_target(username, **kwargs)
    return _post_request(server_url=server_url,
                         resource_path='friends/ids',
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           count=_USERS_COUNT,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC)
def followers(server_url: str,
              username: str='',
              password: str='',
              **kwargs) -> List[int]:
    """Returns IDs of users the auntenticated or
    specified user is followed by (otherwise known as their "friends").

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:param int count: (optional) {count}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:rtype: list
:return: a list of user IDs
    """
    _check_user_target(username, **kwargs)
    return _post_request(server_url=server_url,
                         resource_path='followers/ids',
                         username=username,
                         password=password,
                         data=kwargs).json()



"""
gnusocial.favorites
~~~~~~~~~~~~~~~~~~~

Module with favorite resources.
"""
from .utils import _post_request, _check_user_target, docstring
from .docs import (_SERVER_URL_DOC, _USERNAME_DOC, _PASSWORD_DOC,
                   _SINCE_ID_DOC, _MAX_ID_DOC, _STATUSES_COUNT, _STATUS_DICT,
                   _STATUS_ID_DOC)


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_STATUSES_COUNT,
           status_dict=_STATUS_DICT)
def favorites(server_url: str,
              username: str,
              password: str,
              **kwargs) -> list:
    """Returns the 20 most recent notices favorited by the authenticating or
    specified user.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:rtype: list
:return: list of dicts with following structure:

::

    {status_dict}
    """
    _check_user_target(**kwargs)
    return _post_request(server_url=server_url,
                         resource_path='favorites',
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           status_id=_STATUS_ID_DOC,
           status_dict=_STATUS_DICT)
def create(server_url: str,
           username: str,
           password: str,
           status_id: int) -> dict:
    """Favorites the status specified in the ID parameter as the
    authenticating user. Returns the liked status when successful.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param int status_id: {status_id}
:rtype: dict
:return: dict with following structure:

::

    {status_dict}
    """
    return _post_request(server_url=server_url,
                         resource_path='favorites/create/%d' % status_id,
                         username=username,
                         password=password).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           status_id=_STATUS_ID_DOC,
           status_dict=_STATUS_DICT)
def destroy(server_url: str,
            username: str,
            password: str,
            status_id: int) -> dict:
    """Unfavorites the status specified in the ID parameter as the
    authenticating user. Returns the unliked status when successful.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
:param int status_id: {status_id}
:rtype: dict
:return: dict with following structure:

::

    {status_dict}
    """
    return _post_request(server_url=server_url,
                         resource_path='favorites/destroy/%d' % status_id,
                         username=username,
                         password=password).json()



"""
gnusocial.friendships
~~~~~~~~~~~~~~~~~~~~~

Module with friendship resources.
"""
from typing import Union
from .utils import _post_request, _check_user_target, docstring
from .docs import (_SERVER_URL_DOC, _USERNAME_DOC, _PASSWORD_DOC, _USER_DICT,
                   _USER_ID_DOC, _SCREEN_NAME_DOC, _SOURCE_USER_DOC,
                   _TARGET_USER_DOC, _SOURCE_ID_DOC, _TARGET_ID_DOC,
                   _SOURCE_SCREEN_NAME_DOC, _TARGET_SCREEN_NAME_DOC,
                   _RELATIONSHIP_DICT)


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           user_dict=_USER_DICT)
def create(server_url: str,
           username: str,
           password: str,
           **kwargs) -> dict:
    """Allows the authenticating users to follow the user specified in the
    user_id or screen_name parameter.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:rtype: dict
:return: dict with following structure:

::

    {user_dict}
    """
    _check_user_target(**kwargs)
    return _post_request(server_url=server_url,
                         resource_path='friendships/create',
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           user_dict=_USER_DICT)
def destroy(server_url: str,
            username: str,
            password: str,
            **kwargs) -> dict:
    """Allows the authenticating users to unfollow the user specified in the
    user_id or screen_name parameter.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:rtype: dict
:return: dict with following structure:

::

    {user_dict}
    """
    _check_user_target(**kwargs)
    return _post_request(server_url=server_url,
                         resource_path='friendships/destroy',
                         username=username,
                         password=password,
                         data=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           source_user=_SOURCE_USER_DOC,
           target_user=_TARGET_USER_DOC)
def exists(server_url: str,
           source_user: Union[str, int],
           target_user: Union[str, int],
           username: str='',
           password: str='') -> bool:
    """Shows if source_user follows target_user.

:param str server_url: {server_url}
:param source_user: User that is following. Can be an ID or screen name.
:type source_user: int or str
:param target_user: User that is followed. Can be an ID or screen name.
:type target_user: int or str
:param str username: (optional) {username}
:param str password: (optional) {password}
:rtype: bool
:return: `True` if `source_user` follows `target_user`. `False` otherwise.
    """
    data = {'user_a': source_user, 'user_b': target_user}
    return _post_request(server_url=server_url,
                         resource_path='friendships/exists',
                         username=username,
                         password=password,
                         data=data).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           source_id=_SOURCE_ID_DOC,
           source_screen_name=_SOURCE_SCREEN_NAME_DOC,
           target_id=_TARGET_ID_DOC,
           target_screen_name=_TARGET_SCREEN_NAME_DOC,
           relationship_dict=_RELATIONSHIP_DICT)
def show(server_url: str,
         username: str='',
         password: str='',
         **kwargs) -> dict:
    """Returns detailed information about the relationship between two
    arbitrary users.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int source_id: (optional) {source_id}
:param str source_screen_name: (optional) {source_screen_name}
:param int target_id: (optional) {target_id}
:param str target_screen_name: (optional) {target_screen_name}
:rtype: dict
:return: dict with following structure:

::

    {relationship_dict}
    """
    has_target_id = 'target_id' in kwargs
    has_target_screen_name = 'target_screen_name' in kwargs
    if has_target_id == has_target_screen_name:
        raise Exception(
            "You must either specify target_id or target_screen_name."
        )
    has_source_id = 'source_id' in kwargs
    has_source_screen_name = 'source_screen_name' in kwargs
    if has_source_id and has_source_screen_name:
        raise Exception(
            "You must either specify source_id or source_screen_name."
        )
    if (not has_source_id and not has_target_screen_name) and not username:
        raise Exception(
            "You must either specify source_id or " +
            "source_screen_name or username")
    return _post_request(server_url=server_url,
                         resource_path='friendships/show',
                         username=username,
                         password=password,
                         data=kwargs).json()



#!/usr/bin/env python3
import codecs
from setuptools import setup
try:
    codecs.lookup('mbcs')
except LookupError:
    def func(name, enc=codecs.lookup('ascii')):
        return {True: enc}.get(name == 'mbcs')
    codecs.register(func)


setup(name='gnusocial',
      version='1.0.0',
      description='GNU Social API for Python 3',
      long_description=open('README.rst').read(),
      author='dtluna',
      author_email='dtluna@openmailbox.org',
      maintainer='dtluna',
      maintainer_email='dtluna@openmailbox.org',
      license='GPLv3',
      classifiers=[
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3',
      ],
      url='https://gitgud.io/dtluna/pygnusocial',
      platforms=['any'],
      packages=['gnusocial'],
      requires=['requests'])



"""
gnusocial.docs
~~~~~~~~~~~~~~

Module with common docstring constants.
"""

_SERVER_URL_DOC = 'URL of the server'
_USERNAME_DOC = 'name of the authenticating user'
_PASSWORD_DOC = 'password of the authenticating user'
_NAME_DOC = 'full name associated with the profile'
_URL_DOC = '''URL associated with the profile.
        Will be prepended with "http://" if not present'''
_FULLNAME_DOC = 'full name associated with the group'
_HOMEPAGE_DOC = 'home page URL associated with the group'
_ALIASES_DOC = 'aliases that group has'

_LOCATION_DOC = '''The city or country describing
        where the %s is located.
        The contents are not normalized or geocoded in any way'''
_USER_LOCATION = _LOCATION_DOC % 'user of the account'
_GROUP_LOCATION = _LOCATION_DOC % 'group'

_DESCRIPTION_DOC = 'A description of the %s'
_USER_DESCRIPTION = _DESCRIPTION_DOC % 'user owning the account'
_GROUP_DESCRIPTION = _DESCRIPTION_DOC % 'group'

_PROFILE_LINK_COLOR_DOC = '''Sets a hex value that controls the
        color scheme of links used on the authenticating user’s profile page.
        This must be a valid hexadecimal value,
        and may be either three or six characters (ex: F00 or FF0000).'''
_SCREEN_NAME_DOC = '''The screen name of the user for whom to
        return results for.'''
_USER_ID_DOC = '''The ID of the user for whom to return results
        for. '''

_COUNT_DOC = '''Specifies the number of %s to try
        and retrieve. If not provided, defaults
        to 20.'''
_USERS_COUNT = _COUNT_DOC % 'users'
_DM_COUNT = _COUNT_DOC % 'direct messages'
_STATUSES_COUNT = _COUNT_DOC % 'statuses'
_GROUP_COUNT = _COUNT_DOC % 'groups'

_MAX_ID_DOC = '''Returns results with an ID less than
        (that is, older than) or equal to the specified ID.'''
_SINCE_ID_DOC = '''Returns results with an ID greater than
        (that is, more recent than) the specified ID.'''
_TEXT_DOC = 'The text of your direct message.'
_STATUS_ID_DOC = 'The numerical ID of the desired status.'
_TARGET_USER_DOC = 'User that is followed. Can be an ID or screen name.'
_SOURCE_USER_DOC = 'User that is following. Can be an ID or screen name.'
_SOURCE_ID_DOC = 'The user_id of the subject user.'
_SOURCE_SCREEN_NAME_DOC = 'The screen_name of the subject user.'
_TARGET_ID_DOC = 'The user_id of the target user.'
_TARGET_SCREEN_NAME_DOC = 'The screen_name of the target user.'
_GROUP_ID_DOC = '''The ID of the group for which to return
        results for.'''
_GROUP_NAME_DOC = '''The name of the group for which to return
        results for.'''
_CONFIG_DICT = """{
        'attachments':
        {
            'file_quota': int, # maximum size of attachment in bytes
            'uploads': bool # True if users are allowed to upload files
        },
        'group':
        {
            'desclimit': int
        },
        'integration':
        {
            'source': str
        },
        'license':
        {
            'image': str,
            'owner': str,
            'title': str,
            'type': str,
            'url': str
        },
        'nickname':
        {
            'featured': list
        },
        'notice':
        {
            'contentlimit': int
        },
        'profile':
        {
            'biolimit': int
        },
        'site':
        {
            'broughtby': str,
            'broughtbyurl': str,
            'closed': bool,
            'email': str,
            'fancy': str,
            'inviteonly': bool,
            'language': str,
            'logo': str,
            'name': str,
            'path': str,
            'private': bool,
            'server': str,
            'ssl': str,
            'sslserver': str,
            'textlimit': str,
            'theme': str,
            'timezone': str
        },
        'throttle':
        {
            'count': int,
            'enabled': bool,
            'timespan': int
        },
        'url':
        {
            'maxnoticelength': int,
            'maxurllength': int
        },
        'xmpp':
        {
            'enabled': bool,
            'port': int,
            'server': str,
            'user': str
        }
    }"""

_USER_DICT = """{
        'background_image': str, # URL to background image or False if none
        'backgroundcolor': str, # background color in hex or False if default
        'cover_photo': str, # URL to cover image or False if none
        'created_at': str, # the date of user registration
        'description': str, # user profile description or False if none
        'favourites_count': int, # the number of notices favorited by user
        'followers_count': int,
        'following': bool, # True if authenticating user is following the user
        'friends_count':int, # the number of users user is following
        'groups_count': int, # the number of group user is member of
        'id': int,
        'is_local': bool, # True if user is from server_url
        'is_sandboxed': bool,
        'is_silenced': bool,
        'linkcolor': str, # link color in hex or False if default
        'location': str, # user's location or False if none
        'name': str, # full name associated with the profile
        'notifications': bool, # True if authenticating user is getting
                               # notifications from user
        'ostatus_uri': str, # URL to user profile
        'profile_background_color':str, # same as backgroundcolor
        'profile_banner_url': str, # same as cover_photo
        'profile_image_url': str, # URL to 48x48 avatar image
        'profile_image_url_https': str. # same as profile_image_url, but with
                                        # HTTPS
        'profile_image_url_original': str, # URL to avatar image in original
                                           # resolution
        'profile_image_url_profile_size': str, # URL to 96x96 avatar image
        'profile_link_color': str, # same as linkcolor
        'protected': bool,
        'rights': # a dict of what user can do
        {
            'delete_others_notice': bool,
            'delete_user': bool,
            'sandbox': bool,
            'silence': bool
        },
        'screen_name': str, # user's handle
        'status': dict, # user's latest status
        'statuses_count': int,
        'statusnet_blocking': bool,
        'statusnet_profile_url': str,
        'time_zone': str,
        'url': str, # URL associated with the profile or False if none
        'utc_offset': str
    }"""

_DM_DICT = '''{
        'created_at': str, # date of message creation
        'id': int,
        'recipient': dict, # user dict
        'recipient_id': int,
        'recipient_screen_name': str,
        'sender': dict, # user dict
        'sender_id': int,
        'sender_screen_name': str,
        'text': str
    }'''

_STATUS_DICT = '''{
        'attachments':
        [
            {
                'height': str,
                'id': str,
                'large_thumb_url': str,
                'mimetype': str,
                'oembed': bool,
                'size': str, # size of attacment in kilobytes
                'thumb_url': str,
                'url': str,
                'version': str,
                'width': str
            }
        ],
        'attentions':
        [
            {
                'fullname': str,
                'id': int,
                'ostatus_uri': str,
                'profileurl': str,
                'screen_name': str
            }
        ],
        'created_at': str,
        'external_url': str,
        'fave_num': int, # number of users favorited the notice
        'favorited': bool, # True if favorited by authenticated user
        'geo': str,
        'id': int,
        'in_reply_to_ostatus_uri': str,
        'in_reply_to_profileurl': str,
        'in_reply_to_screen_name': str,
        'in_reply_to_status_id': str,
        'in_reply_to_user_id': int,
        'is_local': bool,
        'is_post_verb': bool,
        'repeat_num': int,
        'repeated': bool, # True if repeated by authenticated user
        'repeated_id': int,
        'retweeted_status': dict, # repeated status dict
        'source': str,
        'statusnet_conversation_id': int,
        'statusnet_html': str, # HTML contents of the notice
        'statusnet_in_groups': bool,
        'text': str, # plain text contents of the notice
        'truncated': bool,
        'uri': str,
        'user': dict # user dict
    }'''

_RELATIONSHIP_DICT = '''{
        'relationship':
        {
            'source':
            {
                'blocking': bool, # True if source user is blocking target user
                'followed_by': bool, # True if source user is followed by
                                     # target user
                'following': bool, # True if source user is following target
                                   # user
                'id': int,
                'notifications_enabled': bool, # If notifications about target
                                               # user are enabled for source
                                               # user
                'screen_name': str
            },
            'target': dict # same as source
        }
    }'''

_GROUP_DICT = '''{
        'admin_count': int,
        'blocked': bool,
        'created': str,
        'description': str,
        'fullname': str,
        'homepage': str,
        'homepage_logo': str,
        'id': int.
        'location': str,
        'member': bool,
        'member_count': int,
        'mini_logo': str,
        'modified': str,
        'nickname': str,
        'original_logo': str,
        'stream_logo': str,
        'url': str
    }'''
